#ifndef __NJAMD_LIB_MEM_H__
# define __NJAMD_LIB_MEM_H__
#include <config.h>
#define NJAMD_INFO PACKAGE VERSION " by Mike Perry\n"

/* Define this to the stack depth you want to trace */
#define TRACE_DEPTH     3

/* various debugging macros */
#ifdef EDEBUG
# define TRACEF 	write(2, __FUNCTION__"\n", strlen(__FUNCTION__)+1)
# define TR			do \
					{ \
						write(2, __FILE__":"__FUNCTION__"\n", strlen(__FILE__":"__FUNCTION__"\n")); \
						func = __FUNCTION__; line = __LINE__; file = __FILE__; \
					} while(0)
# define PRINTF(x...)	__nj_eprintf( __FUNCTION__ ": " ##x)
# define EPRINTF(x...)  __nj_eprintf( __FUNCTION__ ": " ##x)
# define __USE_UNIX98
#elif DEBUG
# define __USE_UNIX98
# define TRACEF 	write(2, __FUNCTION__"\n", strlen(__FUNCTION__)+1)
# define TR		do { file = __FILE__; func = __FUNCTION__; line = __LINE__; } while(0)
# define PRINTF(x...)	__nj_eprintf( __FUNCTION__ ": " ##x)
# define EPRINTF(x...)
#else
# define TRACEF
# define TR
# define PRINTF(x...)
# define EPRINTF(x...) 
#endif
	
#include <stdio.h>
#ifdef _THREAD_SAFE
# include <pthread.h>
#endif
#include <unistd.h>
#include <sys/param.h> /* stupid openbsd */
#include <limits.h>
#include <sys/types.h>
#include <signal.h>
#ifdef HAVE_SYS_USER_H
# include <sys/user.h>
#endif
#include <sys/mman.h>

/* Constants */
#ifndef TRUE
# define TRUE				1
# define FALSE				0
#endif

/* Methods of protecting freed memory */
#define ERROR_CHK_FREE		0
#define SEGV_CHK_FREE		1
#define NO_CHK_FREE			2
#define NO_FREE				3

/* Allocation methods */
#define ALLOC_OVER			0
#define ALLOC_UNDER			1
#define ALLOC_SUNDER		2
#define ALLOC_NONE			3

#define LINUX_PROC_OVERCOMMIT		"/proc/sys/vm/overcommit_memory"
#define LINUX_PROC_MAX_MAP_COUNT	"/proc/sys/vm/max_map_count"

/* The value to fencepost our buffers with to check for consistancy */
#define NJ_FENCEPOST			(0xdeadbeef)

/* Dummy return addresses */ 
#define NJ_DESTRUCTOR			(-1)
#define NJ_RETADDR_NONE			(-3)
#define NJ_RETADDR_NOTFREE		(-5)
	
/* Try to find NJ_PAGE_SIZE */
#if defined(HAVE_PAGE_SIZE)
# ifndef NJ_PAGE_SIZE
#  if defined(PAGE_SIZE)
#   define NJ_PAGE_SIZE PAGE_SIZE
#  elif defined(EXEC_PAGESIZE)
#   define NJ_PAGE_SIZE EXEC_PAGESIZE
#  elif defined(NBPG)
#   ifndef CLSIZE
#    define CLSIZE 1
#   endif
#   define NJ_PAGE_SIZE (NBPG*CLSIZE)
#  elif defined(NBPC)
#   define NJ_PAGE_SIZE (NBPC)
#  elif defined(PAGESIZE)
#   define NJ_PAGE_SIZE PAGESIZE
#  endif
# endif
#endif

#if defined(HAVE_PAGE_SHIFT)
# define NJ_PAGE_SHIFT PAGE_SHIFT
#endif

#if defined(HAVE_PAGE_MASK)
# ifdef NEGATE_PAGE_MASK
#  define NJ_PAGE_MASK			(~PAGE_MASK)
# else
#  define NJ_PAGE_MASK 			  PAGE_MASK
# endif
#endif


/* MAP ANONYMOUS */ 
#if !defined(MAP_ANONYMOUS) && defined(MAP_ANON)              
# define MAP_ANONYMOUS MAP_ANON                               
#endif                                                        
                                                              
#if defined(MAP_ANONYMOUS) && !defined(MAP_ANON)              
# define MAP_ANON MAP_ANONYMOUS                               
#endif          
	
#ifndef MAP_ANON
# define MAP_ANON 0
# define MAP_ANONYMOUS 0
#endif

#define entry_is_freed(entry) \
	((entry)->loc.freed[1] != NJ_RETADDR_NOTFREE)

/* Also in fe/njamd.h */
#ifndef HAVE_VSNPRINTF
# define vsnprintf(s, n, f, ap)			vsprintf(s, f, ap)
#endif

#ifndef HAVE_SNPRINTF
# define snprintf(s, n, f, ap...)		sprintf(s, f, ## ap)
#endif

/* Alignment macros */
	
/* Align to a certain number of bytes */
#define ALIGN_TO(val, alignto)				\
		((alignto > 1) ? ((u_long)(val) + (alignto)) & (~((alignto) - 1)) \
						: (val))

/* Aligned objects should lack the lower bits */
#define IS_ALIGNED(val, alignto)		\
		!((u_long)(val) & ((alignto) - 1))

/* Overhead due to aligning */
#define ALIGN_OVERHEAD(val, alignto)	\
		(ALIGN_TO(val, alignto) - (u_long)(val))

#define NJ_MIN(a,b)			((a) < (b) ? (a) : (b))
#define NJ_MAX(a,b)			((a) > (b) ? (a) : (b))

#define nj_exit(ret)				\
	do { 							\
		if(NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))	\
		{	\
			*((struct mem_stats *)__nj_heap_top) = __nj_mem_stats; \
			msync((void *)__nj_heap_tbl, (u_long)__nj_heap_top - (u_long)__nj_heap_tbl + sizeof(struct mem_stats), \
					MS_SYNC);\
			truncate(__nj_heap_file, (u_long)__nj_heap_top - (u_long)__nj_heap_tbl + sizeof(struct mem_stats));\
		}\
		else if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY)) \
			unlink(__nj_heap_file); \
		_exit(ret); \
	} while(0);

	
/* Data types */
struct mem_stats
{
	u_int total_user; /* Total user memory ever requested */
	u_int total_faulted; /* Total memory faulted by system (ie pages present) */
	u_int peak_user; 
	u_int peak_faulted;
	u_int peak_overhead;
	u_int total_allocs; /* Total number of allocation requests */
	u_int leaked_allocs; /* Total number of leaked allocs */
	u_long as_used;		/* Total amount of address space used */
	u_char alloc_type; /* Allocation type */
};

struct slab_desc
{
	void *addr; /* address of the slab */
	void *next; /* start of next chunk of memory in the slab to slice out */
	size_t remainder;
};

struct slab_tbl
{
	int size;
	struct slab_desc *tbl;
};

struct pthread_arg_wrapper
{
	void *(*start)(void *);
	void *real_arg;
};

struct called 
{
	u_long alloced[TRACE_DEPTH];
	u_long freed[TRACE_DEPTH];
};

/* Heap entry. Describes both an open slot, and an occupied one */
struct heap_entry
{
	union addr_u {
		void *start; /* Location of buffer in memory */
		struct heap_entry *next; /* next free slot (used to make a stack) */
	} u_addr;
	struct called loc; /* where we were alloc'ed / freed */
	size_t ulen; /* Length of the user segment */
};

typedef RETSIGTYPE (*nj_sighandler_t)(int);
#ifdef HAVE_POSIX_SIGACTION
typedef void (*nj_sigaction_t)(int, siginfo_t *, void *);
#endif
typedef void *(*nj_allocator_t)(void *, size_t, int, u_long *, char *);
typedef void (*nj_deallocator_t)(void *, u_long *);

/* Prototypes */
	
/* mem_xx.c */
void *__nj_none_alloc		(void *, size_t, int, u_long *, char *);
void *__nj_overflow_alloc	(void *, size_t, int, u_long *, char *);
void *__nj_sunderflow_alloc	(void *, size_t, int, u_long *, char *);
void *__nj_underflow_alloc	(void *, size_t, int, u_long *, char *);
void __nj_none_free			(void *, u_long *);
void __nj_overflow_free		(void *, u_long *);
void __nj_sunderflow_free	(void *, u_long *);
void __nj_underflow_free	(void *, u_long *);
	
/* setup.c */
void __nj_primary_init		(void);
void __nj_secondary_init	(void);
void __nj_register_handlers	(void);

/* util.c */
struct heap_entry *__nj_new_heap_entry	(void *, int, u_long *);
void __nj_free_stack_push				(struct heap_entry *);
void __nj_eprintf						(const char *, ...);
void __nj_print_stack_trace				(u_long *, int);
void __nj_update_stats					(int, int, int);
void __nj_perror						(const char *);
void *__nj_slab_fetch					(size_t);
int  __nj_address_from_malloc			(void *);
char *__nj_ptoa							(void *);
void *__nj_pthread_launch				(void *);
void __nj_ptr_info						(u_long);
void __nj_pthread_launch_end			(void);
struct heap_entry *__nj_free_stack_pop	(void);
void __nj_userspace_ret					(u_long *, int);
int  __nj_entered_usercode				(void);
void __nj_prot_to_none					(void);
void __nj_out_of_memory					(void);
void __nj_usage_stats					(void);
void __nj_dump_leaks					(int);
int  __nj_slab_alloc					(int);
char *__nj_itoa							(int);
	
/* public.c */
void __nj_pthread_exit_start			(void);
void __nj_pthread_exit_end				(void);
void __nj_exit_start					(void);
void __nj_exit_end						(void);

/* Global variables */
#if defined(DEBUG) || defined(EDEBUG)
extern char 					*file;
extern char 					*func;
extern int 						line;
#endif

/* global mm data structures */
extern struct heap_entry 		*__nj_heap_stack;
extern struct heap_entry 		*__nj_heap_tbl;
extern struct heap_entry 		*__nj_heap_top;
#ifdef _THREAD_SAFE
extern pthread_mutex_t 			__nj_pthread_lock;
extern pthread_mutex_t 			__nj_heap_lock;
extern pthread_mutex_t 			__nj_slab_lock;
#endif
extern struct slab_tbl	 		__nj_slab_tbl;
extern int 						__nj_zfd;
extern int 						__nj_efd;
#ifdef NJ_PAGE_SIZE
extern u_char 					__nj_zero_page[NJ_PAGE_SIZE];
#else
extern u_char					*__nj_zero_page;
#endif
extern int						__nj_anonfd;


/* Allocator functions */
extern nj_allocator_t			__nj_known_alloc;
extern nj_deallocator_t			__nj_known_free;

/* Prefrence variables */
extern int						__nj_display_depth;
extern int 						__nj_known_chk_free;
extern u_long 					__nj_align;
extern int						__nj_prot;

/* Status variables */
extern char		 				__nj_heap_file[80];
extern u_long					__nj_known_slab0;
extern u_long					__nj_slab0;
extern u_long					__nj_sbrk0;
extern struct mem_stats			__nj_mem_stats;
extern u_int 					__nj_leaked_user;
extern u_int 					__nj_leaked_faulted;

/* Flags */
#define NJ_SET(var, fl)			(var) |= (fl)
#define NJ_ISSET(var, fl)		((var) & (fl))
#define NJ_CLEAR(var, fl)		(var) &= ~(fl)

/* Prefrence flags */
#define NJ_NO_RETADDRS			(1<<0)
#define NJ_PERSIST_HEAP 		(1<<1)
#define NJ_STANDALONE			(1<<2)
#define NJ_TRACE_LIBS			(1<<3)
#define NJ_ALLOW_MALLOC0		(1<<4)
#define NJ_ALLOW_FREE0			(1<<5)
#define NJ_DUMP_STATS			(1<<6)
#define NJ_DUMP_HARDCORE		(1<<7)
#define NJ_DUMP_SOFTCORE		(1<<8)

/* Status flags */
#define NJ_RAN_PRIMARY 			(1<<26)
#define NJ_RAN_SECONDARY		(1<<27)
#define NJ_RAN_LIBC				(1<<28)
#define NJ_ENTERED_MAIN			(1<<29)
#define NJ_ACTIVE_PTHREADS		(1<<30)
#define NJ_ENTERED_DESTRUCT		(1<<31)
extern u_int					__nj_flags;

/* old handlers */
extern nj_sighandler_t			__nj_sig_hdlr[sizeof(u_long)*8];
#ifdef HAVE_POSIX_SIGACTION
extern nj_sigaction_t			__nj_sig_actn[sizeof(u_long)*8];
#endif

/* Portability */
#ifndef NJ_PAGE_SIZE
extern u_long 					NJ_PAGE_SIZE;
#endif

#ifndef NJ_PAGE_MASK
# ifdef NJ_PAGE_SIZE
#  define NJ_PAGE_MASK			(~(NJ_PAGE_SIZE-1))
# else
extern u_long 					NJ_PAGE_MASK;
# endif
#endif

#ifndef NJ_PAGE_SHIFT
extern u_long 					NJ_PAGE_SHIFT;
#endif


/* Macros that depend on data types or variables */

#define NJ_ACCOUNTING_OVERHEAD		(sizeof(struct heap_size *))

/* Can't offset by size_t if aligment is DOUBLE */
#define NJ_OFFSET (ALIGNMENT > NJ_ACCOUNTING_OVERHEAD ? \
					ALIGNMENT : NJ_ACCOUNTING_OVERHEAD)

/* The heap table must accomodate enough heap_entries to describe the address
 * space. Worst case is if each heap_entry describes only 2 pages of memory 
 * This is approximately 16 megs on a 32bit address space. Don't worry, it's
 * only 16 megs of ADDRESS SPACE.. it's not used until faulted, and is synced
 * often, so not much of it is resident */
#define HEAP_TBL_SIZE				((((u_long)ADDRESS_SIZE) >> (NJ_PAGE_SHIFT+1)) \
										* sizeof(struct heap_entry) + 1)

/* Round len up to nearest page boundry */
#define ROUND_TO_PAGE(len)			((((len)-1) & NJ_PAGE_MASK) + NJ_PAGE_SIZE)

/* Round len with accounting info to a blocksize usable in NJAMD */
#define ROUND_TO_BLOCK(len)			((((len)+NJ_ACCOUNTING_OVERHEAD-1) \
										& NJ_PAGE_MASK) + (2<<NJ_PAGE_SHIFT))

/* Round len with accounting info to a page boundary */
#define ROUND_TO_FAULT(len)			((((len)+NJ_ACCOUNTING_OVERHEAD-1) \
										& NJ_PAGE_MASK) + NJ_PAGE_SIZE)

#define SHIFT_MEGABYTE		20	
#define SLAB_MEGS			20
/* FIXME: for 0.9.0 this has to be 64bit */
#define ADDRESS_SIZE		(0xffffffff)
/* Size of a slab (20 megs default) */
#define SLAB_SIZE			(SLAB_MEGS << SHIFT_MEGABYTE)
/* # of entries needed in the slab table to describe the address space
 * (less than 900 bytes total on 32bit procs */
#define SLAB_TABLE_SIZE		ROUND_TO_PAGE((((u_long)ADDRESS_SIZE) / SLAB_SIZE + 1)*sizeof(struct slab_desc))
#define SLAB_TABLE_NENTRIES	(SLAB_TABLE_SIZE/sizeof(struct slab_desc))

/* Settings to use before we can read anything from the environment. 
 * Don't set this to no_alloc */
#define PREINIT_ALLOCATOR		__nj_overflow_alloc
#define PREINIT_DEALLOCATOR		__nj_overflow_free
/* Don't change this to NO_CHK_FREE */
#define PREINIT_CHK_FREE		SEGV_CHK_FREE

/* Default settings. These can be anything */	
#define DEFAULT_ALLOCATOR		__nj_overflow_alloc
#define DEFAULT_DEALLOCATOR		__nj_overflow_free
#define DEFAULT_CHK_FREE		SEGV_CHK_FREE
#define DEFAULT_ALLOC_TYPE		ALLOC_OVER /* Must match ALLOCATOR */

#define NJ_LIMIT_SOFTCORE		(4<<SHIFT_MEGABYTE)

#endif /* __NJAMD_LIB_MEM_H__ */

// vim:ts=4
