#include <lib/njamd.h>
#include <lib/libc.h>
#ifdef HAVE_SANE_DLOPEN
# include <dlfcn.h>
#endif
#include <stdio.h>

nj_sighandler_t (*__njLibcSignal)(int, nj_sighandler_t);
#ifdef HAVE_POSIX_SIGACTION
int (*__njLibcSigaction)(int, const struct sigaction *, struct sigaction *);
#endif

#ifdef _THREAD_SAFE
int  (*__nj_pthread_create)(pthread_t  *  thread, pthread_attr_t *attr, 
		void * (*start_routine)(void *), void *);
void (*__nj_pthread_exit)(void *);
#endif
void (*__nj_exit)(int);

void (*__nj_libc_free)(void *);
void *(*__nj_libc_malloc)(size_t);
void *(*__nj_libc_calloc)(size_t, size_t);
void *(*__nj_libc_realloc)(void *,size_t);

const char *__njLibcFuncInit(void)
{
#ifdef HAVE_SANE_DLOPEN
	void *libc_handle;
# ifdef _THREAD_SAFE
	void *pthread_handle;
# endif
	const char* error = NULL;
#endif
	if (NJ_ISSET(__nj_flags, NJ_RAN_LIBC))
		return NULL;
	TR;
	
#ifdef HAVE_SANE_DLOPEN
	
#ifdef _THREAD_SAFE
	pthread_handle = dlopen(LIBPTHREAD_PATH, RTLD_LAZY);
#endif

	TR;
	/* If we have a libpthread, then standard libc functions are threadsafe.
	   If not, then we get the libc functions from the threadsafe lib if we're
	   building threadsafe version, or the standard path if not. Currently we
	   never define _THREAD_SAFE */
#ifdef HAVE_LIBPTHREAD
	if((libc_handle = dlopen(LIBC_SO_PATH, RTLD_LAZY)) == NULL)
	{
		__nj_eprintf("libc.so not found\n");
		return dlerror();
	}
#elif defined(_THREAD_SAFE) 
	if((libc_handle = dlopen(LIBC_R_PATH, RTLD_LAZY)) == NULL)
	{
		__nj_eprintf("libc.so not found\n");
		return dlerror();
	}
#else
	if((libc_handle = dlopen(LIBC_SO_PATH, RTLD_LAZY)) == NULL)
	{
		__nj_eprintf("libc.so not found\n");
		return dlerror();
	}
#endif

	TR;
	if((__njLibcSignal = dlsym(libc_handle, "signal")) == NULL)
	{
		error = dlerror();
		goto done;
	}
#ifdef HAVE_POSIX_SIGACTION
	if((__njLibcSigaction = dlsym(libc_handle, "sigaction")) == NULL)
	{
		error = dlerror();
		goto done;
	}
#endif

	if((__nj_exit = dlsym(libc_handle, "exit")) == NULL)
	{
		error = dlerror();
		goto done;
	}

	if((__nj_libc_malloc = dlsym(libc_handle, "malloc")) == NULL)
	{
		error = dlerror();
		goto done;
	}

	if((__nj_libc_calloc = dlsym(libc_handle, "calloc")) == NULL)
	{
		error = dlerror();
		goto done;
	}
	
	if((__nj_libc_realloc = dlsym(libc_handle, "realloc")) == NULL)
	{
		error = dlerror();
		goto done;
	}
	
	if((__nj_libc_free = dlsym(libc_handle, "free")) == NULL)
	{
		error = dlerror();
		goto done;
	}

#ifdef _THREAD_SAFE
	if((__nj_pthread_create = dlsym(pthread_handle, "pthread_create")) == NULL)
	{
		error = dlerror();
		goto done;
	}

	if((__nj_pthread_exit = dlsym(pthread_handle, "pthread_exit")) == NULL)
	{
		error = dlerror();
		goto done;
	}
#endif

	NJ_SET(__nj_flags, NJ_RAN_LIBC);
	
	TR;
done:
	dlclose(libc_handle);
#ifdef _THREAD_SAFE
	if(pthread_handle)
		dlclose(pthread_handle);
#endif
	return error;
#endif
	__njLibcSignal = signal;
	return NULL;
}
// vim:ts=4
