/* Util.c - Contains system utility fcns 
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 * 
 */
#include <lib/njamd.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef HAVE_EXECINFO_H
# include <execinfo.h>
#endif

extern char **environ;

/* We must check the frame address as a hack... Sometimes high optimizations  
 * kill the frame pointer even if you don't specify -fomit-frame-pointer..
 * We wouldn't have to do this if the gcc people just followed their
 * documentation and returned NULL instead of segfaulting. I've reported this
 * bug and got no replies. *shrug*
 */
#if defined(USE_FRAME_ADDR_HACK)
# define RA(x)	case x: return (u_long)(__builtin_frame_address(x) ? __builtin_return_address(x) : 0)
#elif !defined(HAVE_DEADLY_RET_ADDR)
# define RA(x)	case x: return (u_long)__builtin_return_address(x)
#else /* __builtin_return_address is not only uselesss, its deadly */
# define RA(x)  case x: return NJ_RETADDR_NONE
#endif

static u_long return_address(int frame)
{
	TR;
	switch(frame+1)
	{
       RA(1);RA(2);RA(3);RA(4);RA(5);RA(6);RA(7);RA(8);RA(9);
	   RA(10);RA(11);RA(12);RA(13);RA(14);RA(15);RA(16);RA(17);RA(18);RA(19);
	   RA(20);RA(21);RA(22);RA(23);RA(24);RA(25);RA(26);RA(27);RA(28);RA(29);
	   RA(30);RA(31);RA(32);RA(33);RA(34);RA(35);RA(36);RA(37);RA(38);RA(39);
	   RA(40);RA(41);RA(42);RA(43);RA(44);RA(45);RA(46);RA(47);RA(48);RA(49);
	   RA(50);RA(51);RA(52);RA(53);RA(54);RA(55);RA(56);RA(57);RA(58);RA(59);
	   RA(60);RA(61);RA(62);RA(63);RA(64);RA(65);RA(66);RA(67);RA(68);RA(69);
	   RA(70);RA(71);RA(72);RA(73);RA(74);RA(75);RA(76);RA(77);RA(78);RA(79);
	   RA(80);RA(81);RA(82);RA(83);RA(84);RA(85);RA(86);RA(87);RA(88);RA(89);
	   RA(90);RA(91);RA(92);RA(93);RA(94);RA(95);RA(96);RA(97);RA(98);RA(99);
	   default:
	       return 0;
	}
	
}

/* Returns true if addr came from malloc */
int __nj_address_from_malloc(void *addr)
{
	int i;
	
	if(__nj_known_chk_free == NO_CHK_FREE || __nj_known_alloc == __nj_none_alloc)
		return 1;
	
	for(i=0; __nj_slab_tbl.tbl[i].addr != NULL; i++)
	{
		if((u_long)__nj_slab_tbl.tbl[i].addr < (u_long)addr 
				&& (u_long)addr < (u_long)__nj_slab_tbl.tbl[i].next)
			return 1;
	}
	return 0;
	
}

/* Returns true if the next return address is not likely to knock us off the
 * call stack. Lower bound is sbrk(0) if we want to stop at main, 0 otherwise */
static int next_address_valid(u_long cmp, u_long lower_bound)
{
	if(!cmp)
		return FALSE;
#ifdef HAVE_SANE_DLOPEN
# ifdef _THREAD_SAFE
	if(NJ_ISSET(__nj_flags, NJ_ACTIVE_PTHREADS))
	{
		/* only continue recursing if we're outside of main AND we're not inside
		 * the NJAMD pthread launching fucntion (which is the end of the call
		 * stack for pthreads */
		if(cmp > lower_bound
				&& !(((u_long)__nj_pthread_launch < cmp && cmp < (u_long)__nj_pthread_launch_end)
					|| ((u_long)__nj_pthread_exit_start < cmp && cmp < (u_long)__nj_pthread_exit_end)
					|| ((u_long)__nj_exit_start < cmp && cmp < (u_long)__nj_exit_end)))
			return TRUE;
	}
	else
# endif
	{
		/* Only continue recursing if we're outside of main and we're not in 
		 * exit */
		if(cmp > lower_bound
				&& !((u_long)__nj_exit_start < cmp && cmp < (u_long)__nj_exit_end))
			return TRUE;
	}
#else
	if(cmp > lower_bound)
		return TRUE;
#endif

	return FALSE;
}

/* Returns the first BUILTIN_STACK_DEPTH address in the call chain.
 * In ELF, user code exists below sbrk(0). (I refer to userspace here
 * meaning static program code, as oppsed to shared lib code).
 */
void __nj_userspace_ret(u_long *rets, int depth)
{
	int i, k;


	/* Three cases for not storing return address must be kept track of, so we
	 * know what to say/do when not displaying them */
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN)) 
	{
		rets[0] = 0;
		return;
	}

	/* This flag isn't used in the library, but it is used in njamdpm */
	if(NJ_ISSET(__nj_flags, NJ_NO_RETADDRS))
	{
		rets[0] = NJ_RETADDR_NONE;
		return;
	}

	TR;

	if(NJ_ISSET(__nj_flags, NJ_ENTERED_DESTRUCT))
	{
		rets[0] = NJ_DESTRUCTOR;
		return;
	}

	rets[0] = (u_long)return_address(1);

	TR;
	
	if(NJ_ISSET(__nj_flags, NJ_TRACE_LIBS))
	{
		/* Only record return addresses where the next one is valid (as in
		 * from library space) */
		for(i = 1; i < depth; i++)
		{
#ifdef HAVE_WORKING_RET_ADDR
			if(next_address_valid(rets[i-1], 0))
#else
			if(next_address_valid(rets[i-1], __nj_sbrk0))
#endif
			{
				rets[i] = (u_long)return_address(i+1);
			}
			else
			{
				for(; i < depth; i++)
					rets[i] = 0;
				break;
			}
		}

	}
	else
	{
		/* Find the first invalid address and trace up from there */
		for(k=1; next_address_valid(rets[0], __nj_sbrk0); k++)
			rets[0] = (u_long)return_address(k+1);

#ifdef HAVE_WORKING_RET_ADDR
		/* fill in all return addresses with valid ones from usercode */
		for(i = 1; i < depth; i++)
		{
			if(next_address_valid(rets[i-1], 0))
			{
				rets[i] = (u_long)return_address(k+1);
				k++;
			}
			else
			{
				for(; i < depth; i++)
					rets[i] = 0;
				break;
			}
		}
#else
		for(i=1; i < depth; i++)
			rets[i] = 0;
#endif
	}

	TR;
	return;	
	
}


/* This function attempts to detect if our execution path has flowed through
 * main yet.. The problem is we don't know when to stop the backtrace..
 * Right now we stop when we find a function that is below this library 
 * yet above the user segment. This is about where __libc_start_main lies.. 
 * very hackish to say the least. ANy other ideas?
 */
int __nj_entered_usercode()
{
#ifndef HAVE_DEADLY_RET_ADDR
	int i=1; /* Start at the return address of the caller of this function */
	u_long raddr;

	TR;
	
	for(;i<100;i++)
	{
		if((raddr = return_address(i)) < (u_long)__nj_primary_init && raddr > __nj_sbrk0)
			return FALSE;
		if(raddr < __nj_sbrk0)
			return TRUE;
	}
	TR;
	return FALSE;
#else
	return TRUE; /* __bultin_return_address is deadly. Punt and cross fingers */
#endif
}

/* It's a new invention in computer science, the heap-stack :)
 * but seriously folks.. 
 * This monstrosity maintains the set of dirty heap entries that can be reused
 * when we reuse our address space. It's a stack because we're dealing with an
 * mmap'ed file, so temporal locality is a plus.
 */
struct heap_entry *__nj_free_stack_pop(void)
{
	struct heap_entry *ret;

#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_heap_lock);
#endif

	if(__nj_heap_stack)
	{
		ret = __nj_heap_stack;
		__nj_heap_stack = __nj_heap_stack->u_addr.next;

#ifdef _THREAD_SAFE
		pthread_mutex_unlock(&__nj_heap_lock);
#endif
		return ret;
	}

#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_heap_lock);
#endif
	return NULL;
}

void __nj_free_stack_push(struct heap_entry *entry)
{
#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_heap_lock);
#endif

	memset(entry, 0, sizeof(struct heap_entry));
	entry->u_addr.next = __nj_heap_stack;
	__nj_heap_stack = entry;

#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_heap_lock);
#endif
}

struct heap_entry *__nj_new_heap_entry(void *start, int len, u_long *retaddr)
{
	struct heap_entry *ret;

	TR;
	/* If we're reusing the address space, then we have to use the list */
	if(!(ret = __nj_free_stack_pop()))
	{
#ifdef _THREAD_SAFE
		pthread_mutex_lock(&__nj_heap_lock);
#endif
	
		ret = __nj_heap_top++;
		memset(__nj_heap_top, 0, sizeof(struct heap_entry));
#ifdef _THREAD_SAFE
		pthread_mutex_unlock(&__nj_heap_lock);
#endif
	}

	TR;

	ret->u_addr.start = start;
	ret->ulen = len;
	ret->loc.freed[1] = NJ_RETADDR_NOTFREE;
	memcpy(ret->loc.alloced, retaddr, TRACE_DEPTH*sizeof(u_long));

	return ret;
	
}


/* Allocates slabs with a size backoff algorithm to compensate for
 * address space fragmentation */
int __nj_slab_alloc(int idx)
{
	int i;

	/* This function needn't be locked because it's only called on init
	 * (before threads exist) or inside another lock */

	/* This is a good time to write the heap table out to disk */
	if(NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))
		msync(__nj_heap_tbl, (u_long)__nj_heap_top - (u_long)__nj_heap_tbl, 
				MS_ASYNC);

	TR;
	
	if(!__nj_slab_tbl.tbl)
	{
		TR;
		if((__nj_slab_tbl.tbl = mmap(NULL,
						SLAB_TABLE_SIZE,
						PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
						__nj_anonfd, 0)) == (void *)-1)
		{
			return -1;
		}
		memset(__nj_slab_tbl.tbl, 0, SLAB_TABLE_SIZE);
		__nj_slab_tbl.size = SLAB_TABLE_NENTRIES;
	}

	/* sucessively back off the requested size till we find something that
	 * fits or were asking for less than 2 pages. */
	for (i=0; (SLAB_SIZE >> i) >= (2 << NJ_PAGE_SHIFT); i++)
	{
		if((__nj_slab_tbl.tbl[idx].addr = mmap(NULL, SLAB_SIZE >> i, 
						PROT_READ|PROT_EXEC|PROT_WRITE, 
						MAP_PRIVATE|MAP_ANON, __nj_anonfd, 0)) != (void *)-1)
		{
			__nj_slab_tbl.tbl[idx].remainder = SLAB_SIZE >> i;
			__nj_slab_tbl.tbl[idx].next = __nj_slab_tbl.tbl[idx].addr;
	
			return 0;
		}
	}
	
	TR;
	return -1;	
}

static int __nj_slab_resize(void)
{
	struct slab_tbl old_tbl = __nj_slab_tbl;

	TR;
	if(!__nj_slab_tbl.tbl)
	{
		TR;
		if((__nj_slab_tbl.tbl = mmap(NULL,
						SLAB_TABLE_SIZE,
						PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
						__nj_anonfd, 0)) == (void *)-1)
		{
			__nj_perror("NJAMD: Can't resize slab table");
			return -1;
		}
		memset(__nj_slab_tbl.tbl, 0, SLAB_TABLE_SIZE);
		__nj_slab_tbl.size = SLAB_TABLE_NENTRIES;
		return 0;
	}

	TR;
	/* double it */	
	__nj_slab_tbl.size <<= 1;
	
	if((__nj_slab_tbl.tbl = mmap(NULL,
					ROUND_TO_PAGE(__nj_slab_tbl.size*sizeof(struct slab_desc)),
					PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
					__nj_anonfd, 0)) == (void *)-1)
	{
		__nj_perror("NJAMD: Can't resize slab table");
		return -1;
	}
	memset(__nj_slab_tbl.tbl, 0,
			ROUND_TO_PAGE(__nj_slab_tbl.size*sizeof(struct slab_desc)));
	__nj_slab_tbl.size = ROUND_TO_PAGE(__nj_slab_tbl.size*sizeof(struct slab_desc))/sizeof(struct slab_desc);

	memcpy(__nj_slab_tbl.tbl, old_tbl.tbl, old_tbl.size*sizeof(struct slab_desc));
	
	munmap(old_tbl.tbl, old_tbl.size*sizeof(struct slab_desc));	

	return 0;
}

/* Fetch a section of memory from a slab or directly from a mmap */
void *__nj_slab_fetch(size_t size)
{
	int i;
	void *buf;

	TR;
#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_slab_lock);
#endif

	for(i = 0; 
			i < __nj_slab_tbl.size 
			&& __nj_slab_tbl.tbl[i].addr != NULL 
			&& __nj_slab_tbl.tbl[i].remainder < size; 
			i++)
		;

	if(size > SLAB_SIZE)
	{
		for(; __nj_slab_tbl.tbl[i].addr != NULL; i++)
			;

		PRINTF("LOTS of memory requested (%d)\n", size);
			
		if((__nj_slab_tbl.tbl[i].addr = mmap(NULL, size, 
						PROT_READ|PROT_EXEC|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
						__nj_anonfd, 0)) == (void *)-1)
			goto err;

		__nj_slab_tbl.tbl[i].remainder = 0;
		__nj_slab_tbl.tbl[i].next = __nj_slab_tbl.tbl[i].addr+size;

		buf = __nj_slab_tbl.tbl[i].addr;
#ifdef _THREAD_SAFE
		/* Kudos to Christian Gruner for catching this. */
		pthread_mutex_unlock(&__nj_slab_lock);
#endif
		
		return buf;		
	}

	TR;

	if(i >= __nj_slab_tbl.size)
	{
		TR;	
		if(__nj_slab_resize())
			goto err;
		
#ifdef _THREAD_SAFE
		pthread_mutex_unlock(&__nj_slab_lock);
#endif
		return __nj_slab_fetch(size);
	}

	TR;	
	if(__nj_slab_tbl.tbl[i].addr == NULL)
	{
		if(__nj_slab_alloc(i) != 0)
		{
			if((buf = mmap(NULL, size, 
							PROT_READ|PROT_EXEC|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
							__nj_anonfd, 0)) == (void *)-1)
				goto err;
	
#ifdef _THREAD_SAFE
			pthread_mutex_unlock(&__nj_slab_lock);
#endif
			return buf;
		}
	}
	
	__nj_slab_tbl.tbl[i].next += size;
	__nj_slab_tbl.tbl[i].remainder -= size;
	
	TR;		
	buf = __nj_slab_tbl.tbl[i].next - size;	
#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_slab_lock);
#endif
	return buf;

err:
	
#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_slab_lock);
#endif
	
	if(errno == ENOMEM)
		__nj_out_of_memory();
	else
		__nj_perror("NJAMD: __nj_slab_fetch");
	nj_exit(errno);
	return NULL; /* Nope */
}


void __nj_print_stack_trace(u_long *addr_array, int depth)
{
	int i;

	/* Change in NJAMDPM */
	if(NJ_ISSET(__nj_flags, NJ_NO_RETADDRS))
	{
		__nj_eprintf("\tCall stack not saved\n");
		return;
	}
	
	for(i=0; i < depth; i++)
	{
		if(addr_array[i] == 0)
			return;
		else if(addr_array[i] != NJ_DESTRUCTOR)
		{
#ifdef HAVE_BACKTRACE_SYMBOLS_FD
			__nj_eprintf("\tcalled from ");
			backtrace_symbols_fd( (void **)&addr_array[i], 1, __nj_efd );
#else
			__nj_eprintf("\tcalled from 0x%lx\n", addr_array[i]);
#endif
		}
		else
			__nj_eprintf("\tcalled from a destructor during program exit\n");
	}
	
}


static int dump_entry(struct heap_entry *entry, int depth)
{

	if(!entry->ulen)
	{
		__nj_eprintf("\n0x%lx-0x%lx: Freed Memory\n", (u_long)entry->u_addr.start,
				(u_long)entry->u_addr.start + (__nj_mem_stats.alloc_type == ALLOC_NONE ? entry->ulen : ROUND_TO_BLOCK(entry->ulen)));
	}

	/* If zero, this allocation came before we entered main */
	if(!entry->loc.alloced[0])
	{
		TR;
		return 0;
	}
			
	__nj_eprintf("\n0x%lx-0x%lx: Aligned len %d\n   Allocation callstack:\n", 
			(u_long)entry->u_addr.start, 
			(u_long)entry->u_addr.start + (__nj_mem_stats.alloc_type == ALLOC_NONE ? entry->ulen : ROUND_TO_BLOCK(entry->ulen)), 
			entry->ulen);
	__nj_print_stack_trace(entry->loc.alloced, depth);
	
	if(!entry_is_freed(entry))
		__nj_eprintf("   Not Freed\n");
	else
	{
		__nj_eprintf("   Freed callstack:\n");
		__nj_print_stack_trace(entry->loc.freed, depth);
	}

	return 1;
}

static struct heap_entry *find_addr(struct heap_entry *tbl, size_t size,
		u_long addr)
{
	int i;

	for(i=0; i < size; i++)
	{
		/* Make sure ulen is not zero */
		/* and if prot = none, then give the ulen, not the rounded amt */
		if(tbl[i].ulen && (u_long)tbl[i].u_addr.start <= addr
				&& addr < (u_long)tbl[i].u_addr.start
				+ ( (__nj_mem_stats.alloc_type == ALLOC_NONE)
					? tbl[i].ulen
					: ROUND_TO_BLOCK(tbl[i].ulen)))
		{   
			return &tbl[i];
		}
	}

	return NULL;
	
}

/* External symbol that can be called from gdb even if you don't have
 * PERSISTANT_HEAP set */
void __nj_ptr_info(u_long ptr)
{
	struct heap_entry *entry;
	
	if((entry = find_addr(__nj_heap_tbl, 
					((u_long)__nj_heap_top - (u_long)__nj_heap_tbl)/sizeof(struct heap_entry), ptr)) == NULL)
	{
		__nj_eprintf("Address 0x%lx not found in heap\n", (u_long)ptr);
		return;
	}
	dump_entry(entry, TRACE_DEPTH);
}

/* This removes the mapped memory in the slab to save space if we're using
 * PROT_NONE  */
void __nj_prot_to_none(void)
{
	/* Unmap the unused section of the slab */
    munmap((void *)__nj_known_slab0, SLAB_SIZE - (__nj_known_slab0 - __nj_slab0));
	/* Unmap the slab table. We won't be using it no mo */
	munmap((void *)__nj_slab_tbl.tbl,
		            ROUND_TO_PAGE(__nj_slab_tbl.size*sizeof(struct slab_desc)));
}

void __nj_dump_leaks(int depth)
{
	struct heap_entry *p;
	int leaked =0;

	if(depth > TRACE_DEPTH)
	{
		__nj_eprintf("Error, depth %d too large. Compiled max is %d\n", depth, TRACE_DEPTH);
		depth = TRACE_DEPTH;
	}

	for(p = __nj_heap_tbl; p < __nj_heap_top; p++)
	{
		if(p->ulen != 0 && !entry_is_freed(p))
		{
			if(dump_entry(p, depth))
				leaked += p->ulen;
		}
	}

}

/* FUNCTION: __nj_update_stats
 * SUMMARY: This function updates the statistics when amt memory is alloced
 * (> 0) or freed (< 0).
 */
void __nj_update_stats(int user, int faulted, int as)
{
	/* No stats for the system stuff, please. Also, don't bother keeping stats
	 * if no one'll ever see them. */
	if((!NJ_ISSET(__nj_flags, NJ_DUMP_STATS)
			   	&& !NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))
			|| !NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN)) 
			return;

#if 0
	if(NJ_ISSET(__nj_flags, NJ_ENTERED_DESTRUCT))
		__nj_eprintf("Warning: free called in destructor. Please mail mikepery@fscked.org\n");
#endif

	__nj_leaked_user += user;
	__nj_leaked_faulted += faulted;
	__nj_mem_stats.as_used += as;

	if(user > 0)
	{
		__nj_mem_stats.total_user += user;
		__nj_mem_stats.total_faulted += faulted;
		__nj_mem_stats.total_allocs++;
		__nj_mem_stats.leaked_allocs++;

		if(__nj_leaked_user > __nj_mem_stats.peak_user)
		{
			__nj_mem_stats.peak_user = __nj_leaked_user;
			__nj_mem_stats.peak_faulted = __nj_leaked_faulted;
		}
		
		if(__nj_leaked_faulted - __nj_leaked_user > __nj_mem_stats.peak_overhead)
			__nj_mem_stats.peak_overhead = __nj_leaked_faulted - __nj_leaked_user;
	}
	else
		__nj_mem_stats.leaked_allocs--;
		
}

static char *__nj_byte_to_string(char *buf, int len, int bytes)
{
	if(bytes > 1<<SHIFT_MEGABYTE)
		snprintf(buf, len, "%.3f mB", bytes/(1024*1024.0));
	else if(bytes > 1024)
		snprintf(buf, len, "%.3f kB", bytes/1024.0);
	else
		snprintf(buf, len, "%d bytes", bytes);

	return buf;
}

/* FUNCTION: __nj_usage_stats
 * SUMMARY: This is just a quick little stat function so I can see how much
 * memory is mapped. 
 */
void __nj_usage_stats(void)
{
	/* Kinda silly, but what the hell. eprintf is much more expensive than
	 * stack space allocation. */
	char buf1[80], buf2[80], buf3[80], buf4[80], buf5[80], buf6[80], buf7[80];

	if(!__nj_mem_stats.total_allocs)
	{
		__nj_eprintf("\nNJAMD Totals:\n\tmalloc() was never called!\n");
		return;
	}

	__nj_eprintf(
			"\nNJAMD totals:\n\n"
			"\tAllocation totals:\t\t%d total, %d leaked\n"
			"\tLeaked User Memory:\t\t%s\n"
			"\tPeak User Memory:\t\t%s\n"
			"\tNJAMD Overhead at peak:\t\t%s\n" 
			"\tPeak NJAMD Overhead:\t\t%s\n" 
			"\tAverage NJAMD Overhead:\t\t%s per alloc\n" 
			"\tAddress space used:\t\t%s\n"
			"\tNJAMD Overhead at exit:\t\t%s\n\n",
			__nj_mem_stats.total_allocs, __nj_mem_stats.leaked_allocs, 
			__nj_byte_to_string(buf1, sizeof(buf1), __nj_leaked_user), 
			__nj_byte_to_string(buf2, sizeof(buf2), __nj_mem_stats.peak_user), 
			__nj_byte_to_string(buf3, sizeof(buf3), __nj_mem_stats.peak_faulted-__nj_mem_stats.peak_user),
			__nj_byte_to_string(buf4, sizeof(buf5), __nj_mem_stats.peak_overhead),
			__nj_byte_to_string(buf5, sizeof(buf5), (__nj_mem_stats.total_faulted-__nj_mem_stats.total_user)/__nj_mem_stats.total_allocs),
			__nj_byte_to_string(buf6, sizeof(buf6), __nj_mem_stats.as_used), 
			__nj_byte_to_string(buf7, sizeof(buf7), __nj_leaked_faulted-__nj_leaked_user)
				);
}

void __nj_eprintf(const char *format, ...)
{
    char outbuf[1024];
    va_list ap;
    int len;
	va_start(ap, format);
	len = vsnprintf(outbuf, sizeof(outbuf), format, ap);
	if(len > 0)
		write(__nj_efd, outbuf, len);
	else
		write(__nj_efd, outbuf, strlen(outbuf));
}

#ifdef _THREAD_SAFE
/* Isn't this fantastic? We need it so we don't fall off the call stack when
 * using pthreads. I'm pretty proud of this little hack actually :) */
void *__nj_pthread_launch(void *arg)
{
	struct pthread_arg_wrapper warg = *(struct pthread_arg_wrapper *)arg;

	/* This lock initiated in our wrapped version of pthread_create, to
	 * protect the static variable passed to us */
	pthread_mutex_unlock(&__nj_pthread_lock);
	TR;
	PRINTF("warg: %p, start: %p, real arg: %p\n", arg, warg.start,
			warg.real_arg);
	return warg.start(warg.real_arg);
}

/* Placeholder to detrmine the size of __nj_pthread_launch */
void __nj_pthread_launch_end(void)
{
	return;
}
#endif

/* Prints an error message then exits if we run out of memory. */
void __nj_out_of_memory(void)
{
#ifdef __linux__
	int fd;
	char did_oc = 0;
	char mmc_str[32];
	int mmc = 0;
		
	if((fd = open(LINUX_PROC_MAX_MAP_COUNT, O_RDONLY) != -1))
	{
		read(fd, mmc_str, sizeof(mmc_str));
		mmc = strtol(mmc_str, NULL, 0);
		close(fd);
	}
	
	if((fd = open(LINUX_PROC_OVERCOMMIT, O_RDONLY)) != -1)
	{
		read(fd, &did_oc, 1);
		did_oc -= '0';
		close(fd);
	}

	switch(__nj_known_chk_free)
	{
		case ERROR_CHK_FREE:
			if(did_oc && mmc > 65536)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nSet NJAMD_CHK_FREE=segv or none\n");
			else if(did_oc)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nDid you try the proc_map patch?\n");
			else
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nRun sysctl -w vm.overcommit_memory=1 as root, or set NJAMD_CHK_FREE=segv or none\n");
			break;
		case SEGV_CHK_FREE:
			if(did_oc && mmc > 65536)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nSet NJAMD_CHK_FREE=none\n");
			else if(did_oc)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nDid you try the proc_map patch?\n");
			else
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nRun sysctl -w vm.overcommit_memory=1 as root, or set NJAMD_CHK_FREE=segv or none\n");
			break;
		default:
			if(did_oc && mmc > 65536)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\n");
			else if(did_oc)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nDid you try the proc_map patch?\n");
			else
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nRun sysctl -w vm.overcommit_memory=1 and try the proc_map patch.\n");
			break;
	}
#else
	switch(__nj_known_chk_free)
	{
		case ERROR_CHK_FREE:
			__nj_eprintf("\nNJAMD: Address space exhaustion.\nSet NJAMD_CHK_FREE=segv or none\n");
			break;
		case SEGV_CHK_FREE:
			__nj_eprintf("\nNJAMD: Address space exhaustion.\nSet NJAMD_CHK_FREE=none\n");
			break;
		default:
			__nj_eprintf("\nNJAMD: Address space exhaustion.\nDid you try telling your OS to overcommit virtual memory?\n");
			break;
	}

#endif
	if(NJ_ISSET(__nj_flags, NJ_DUMP_STATS))
		__nj_usage_stats();
	/* Turn off printing of usage stats here */
	nj_exit(ENOMEM);

}

/* FIXME: the following functions are NOT threadsafe.. They are just error
 * reporting functions though. There really is no danger, in that they are
 * only called when something is ALREADY fucked up */
char *__nj_itoa(int i)
{
	static char a[16];
	snprintf(a, 15, "%d", i);
	return a;
}

char *__nj_ptoa(void *i)
{
	static char a[16];
	snprintf(a, 15, "0x%lx", (u_long)i);
	return a;
}

/* We needed our own printf and perror that write immeditaly, and can be
 * guaranteed not to use malloc ever */
void __nj_perror(const char *s)
{
	__nj_eprintf("%s: %s\n", s, strerror(errno));
}

// vim:ts=4
